/* $Id: proto_fics.h,v 1.33 2002/03/17 20:51:47 bergo Exp $ */

/*

    eboard - chess client
    http://eboard.sourceforge.net
    Copyright (C) 2000-2001 Felipe Paulo Guazzi Bergo
    bergo@seul.org

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

// don't #include this, #include protocol.h instead

#ifndef PROTO_FICS_H
#define PROTO_FICS_H 1

#include "stl.h"
#include "position.h"

class FicsProtocolParser : public ProtocolParser {
 public:
  FicsProtocolParser();
  ~FicsProtocolParser();

  void receiveString(char *netstring);
  void sendMove(int x1,int y1,int x2,int y2,int prom);
  void sendDrop(piece p,int x,int y);
  void finalize();
  int  hasAuthenticationPrompts();

  void resign();
  void draw();
  void adjourn();
  void abort();

  void discardGame(int gameid);
  void queryGameList(GameListConsumer *glc);
  void queryAdList(GameListConsumer *glc);
  void observe(int gameid);
  void answerAd(int adid);

  void exaForward(int n);
  void exaBackward(int n);

  void refreshSeeks();

  vector<string *> * getPlayerActions();
  vector<string *> * getGameActions();

  void callPlayerAction(char *player, string *action);
  void callGameAction(int gameid, string *action);
 
 private:
  void createGame(ExtPatternMatcher &pm);
  void createExaminedGame(int gameid,char *whitep,char *blackp);
  void createScratchPos(int gameid,char *whitep,char *blackp);
  void addGame(ExtPatternMatcher &pm);
  void removeGame(ExtPatternMatcher &pm);
  void innerRemoveGame(int gameid);
  void parseStyle12(char *data);
  void gameOver(ExtPatternMatcher &pm, GameResult result);
  void updateGame(ExtPatternMatcher &pm);
  void attachHouseStock();

  void clearMoveBuffer();
  void retrieve1(ExtPatternMatcher &pm);
  void retrieve2(ExtPatternMatcher &pm);
  void retrievef();
  void retrieveMoves(ExtPatternMatcher &pm);

  void sendGameListEntry();
  void sendAdListEntry(char *pat);

  char *knowRating(char *player);
  void clearRatingBuffer();

  void ensureSeekGraph();
  void seekAdd(ExtPatternMatcher &pm);
  void seekRemove(ExtPatternMatcher &pm);

  void doPlayerAction(char *player, int id);
  void doGameAction(int gameid, int id);

  int AuthenticationGone;
  bool LectureBotSpecial;

  PatternBinder Binder;

  PatternMatcher WelcomeMessage;
  PatternMatcher UserName;

  ExtPatternMatcher Login;
  ExtPatternMatcher Password;
  ExtPatternMatcher Prompt;

  ExtPatternMatcher CreateGame;
  ExtPatternMatcher CreateGame0;
  ExtPatternMatcher ContinueGame;
  ExtPatternMatcher BeginObserve;
  ExtPatternMatcher EndObserve;
  ExtPatternMatcher NotObserve;
  ExtPatternMatcher EndExamine;

  ExtPatternMatcher GameStatus;
  ExtPatternMatcher ExaGameStatus;
  ExtPatternMatcher GameListEntry;
  ExtPatternMatcher EndOfGameList;

  ExtPatternMatcher AdListEntry;
  ExtPatternMatcher EndOfAdList;

  ExtPatternMatcher Style12;
  ExtPatternMatcher Style12House;

  ExtPatternMatcher WhiteWon;
  ExtPatternMatcher BlackWon;
  ExtPatternMatcher Drawn;
  ExtPatternMatcher Interrupted;
  ExtPatternMatcher DrawOffer;

  ExtPatternMatcher PrivateTell;
  ExtPatternMatcher News;

  ExtPatternMatcher MovesHeader;
  ExtPatternMatcher MovesHeader2;
  ExtPatternMatcher MovesHeader3;
  ExtPatternMatcher MovesHeader4;
  ExtPatternMatcher MovesBody;
  ExtPatternMatcher MovesBody2;
  ExtPatternMatcher MovesEnd;

  ExtPatternMatcher Kibitz;
  ExtPatternMatcher Whisper;
  ExtPatternMatcher Say;
  ExtPatternMatcher Say2;
  ExtPatternMatcher Shout;
  ExtPatternMatcher cShout;
  ExtPatternMatcher It;
  ExtPatternMatcher ChannelTell;
  ExtPatternMatcher Seek;
  ExtPatternMatcher Notify;
  ExtPatternMatcher Challenge;

  ExtPatternMatcher LectureBotXTell;
  ExtPatternMatcher LectureBotXTellContinued;

  ExtPatternMatcher PTell;

  // seek graph
  ExtPatternMatcher SgClear;
  ExtPatternMatcher SgAdd;
  ExtPatternMatcher SgAdd2;
  ExtPatternMatcher SgRemove;

  int LastWasStyle12;

  char nick[64];
  int LastWasPrompt;
  piece style12table[256];
  int LastColor;
  int LastWasCTell;
  int PendingStatus;
  int RetrievingMoves;
  int RetrievingGameList;
  int RetrievingAdList;

  list<Position *> MoveBuf;
  deque<int> MoveRetrievals;
  vector<string *> PActions, GActions;

  GameListConsumer *lastGLC, *lastALC;

  // 2-position circular buffer to keep rating parsing state
  char xxplayer[2][64];
  char xxrating[2][64];
  int xxnext;
};

#endif
